from openai import OpenAI

from rag4p.integrations.openai import MODEL_GPT4_TURBO
from rag4p.rag.generation.chat.chat_prompt import ChatPrompt
from rag4p.rag.generation.quality.answer_quality_service import AnswerQualityService
from rag4p.rag.tracker.rag_observer import RAGObserver


class OpenAIAnswerQualityService(AnswerQualityService):

    def __init__(self, openai_api_key: str, openai_model: str = MODEL_GPT4_TURBO):
        self.openai_client = OpenAI(
            api_key=openai_api_key,
        )
        self.openai_model = openai_model

    def obtain_answer_to_question_quality(self, chat_prompt: ChatPrompt, rag_observer: RAGObserver):
        response = self.openai_client.chat.completions.create(
            model="gpt-4o",
            messages=[
                {"role": "system",
                 "content": chat_prompt.create_system_message(params={})},
                {"role": "user",
                 "content": chat_prompt.create_user_message(params={
                     "question": rag_observer.question,
                     "answer": rag_observer.answer})},
            ],
            stream=False,
        )

        return response.choices[0].message.content

    def obtain_answer_from_context_quality(self, chat_prompt: ChatPrompt, rag_observer: RAGObserver):
        response = self.openai_client.chat.completions.create(
            model="gpt-4o",
            messages=[
                {"role": "system",
                 "content": chat_prompt.create_system_message(params={})},
                {"role": "user",
                 "content": chat_prompt.create_user_message(params={
                     "context": rag_observer.context,
                     "answer": rag_observer.answer})},
            ],
            stream=False,
        )
        return response.choices[0].message.content


PROVIDER = "bedrock"

MODEL_TITAN_EXPRESS = "amazon.titan-text-express-v1"
EMBEDDING_MODEL_TITAN = "amazon.titan-embed-text-v1"

DEFAULT_MODEL = MODEL_TITAN_EXPRESS
DEFAULT_EMBEDDING_MODEL = EMBEDDING_MODEL_TITAN



from rag4p.integrations.openai.quality.openai_answer_quality_service import OpenAIAnswerQualityService
from rag4p.rag.tracker.rag_observer import RAGObserver
from rag4p.util.key_loader import KeyLoader

if __name__ == '__main__':
    from dotenv import load_dotenv

    load_dotenv()

    question = "Since when was the Vasa available for the public to visit?"
    answer = "The Vasa ship was available for the public to visit since Friday 16 February 1962."
    context = ("By Friday 16 February 1962, the ship is ready to be displayed to the general public at the "
               "newly-constructed Wasa Shipyard, where visitors can see Vasa while a team of conservators, "
               "carpenters and other technicians work to preserve the ship.")

    rag_observer = RAGObserver(question=question, answer=answer, context=context)

    key_loader = KeyLoader()
    answer_quality_service = OpenAIAnswerQualityService(openai_api_key=key_loader.get_openai_api_key())

    quality = answer_quality_service.determine_quality_answer_related_to_question(rag_observer=rag_observer)
    print(f"Quality: {quality.quality}, Reason: {quality.reason}")

    quality = answer_quality_service.determine_quality_answer_from_context(rag_observer=rag_observer)
    print(f"Quality: {quality.quality}, Reason: {quality.reason}")


from typing import List

from rag4p.indexing.indexing_service import IndexingService
from rag4p.integrations.openai import MODEL_GPT4_TURBO
from rag4p.integrations.openai.openai_embedder import OpenAIEmbedder
from rag4p.integrations.openai.openai_question_generator import OpenAIQuestionGenerator
from rag4p.indexing.input_document import InputDocument
from rag4p.indexing.splitters.sentence_splitter import SentenceSplitter
from rag4p.rag.generation.question_generator_service import QuestionGeneratorService
from rag4p.rag.retrieval.quality.retrieval_quality_service import obtain_retrieval_quality
from rag4p.rag.store.local.internal_content_store import InternalContentStore

from rag4p.rag.retrieval.quality.question_answer_record import QuestionAnswerRecord
from rag4p.util.key_loader import KeyLoader

if __name__ == '__main__':
    from dotenv import load_dotenv

    load_dotenv()

    key_loader = KeyLoader()
    embedder = OpenAIEmbedder(api_key=key_loader.get_openai_api_key())
    content_store = InternalContentStore(embedder=embedder)
    doc_text = ("The Vasa was a Swedish warship built between 1626 and 1628. The ship foundered and sank after sailing "
                "about 1,300 m (1,400 yd) into its maiden voyage on 10 August 1628.")
    indexing_service = IndexingService(content_store=content_store)
    indexing_service.index_document(InputDocument(document_id="vasa", text=doc_text, properties={}),
                                    splitter=SentenceSplitter())

    question_generator = OpenAIQuestionGenerator(openai_api_key=key_loader.get_openai_api_key(),
                                                 openai_model=MODEL_GPT4_TURBO)
    question_generator_service = QuestionGeneratorService(retriever=content_store,
                                                          question_generator=question_generator)

    question_answer_pairs = []  # type: List[QuestionAnswerRecord]
    for chunk in content_store.loop_over_chunks():
        # Use LLM to generate a question for this chunk
        question = question_generator_service.generate_question(chunk.chunk_text)

        question_answer_pairs.append(QuestionAnswerRecord(document_id=chunk.document_id,
                                                          chunk_id=chunk.chunk_id,
                                                          chunk_text=chunk.chunk_text,
                                                          question=question))

    retriever_quality = obtain_retrieval_quality(question_answer_records=question_answer_pairs, retriever=content_store)

    print(f"Quality using precision: {retriever_quality.precision()}")
    print(f"Total questions: {retriever_quality.total_questions()}")



import os
from abc import ABC, abstractmethod

from openai import OpenAI

from rag4p.rag.generation.chat.chat_prompt import ChatPrompt
from rag4p.rag.generation.quality.answer_from_context_quality import AnswerFromContextQuality
from rag4p.rag.generation.quality.answer_quality import AnswerQuality
from rag4p.rag.generation.quality.answer_to_question_quality import AnswerToQuestionQuality
from rag4p.rag.tracker.rag_observer import RAGObserver


class AnswerQualityService(ABC):

    def determine_quality_of_answer(self, rag_observer: RAGObserver) -> AnswerQuality:
        answer_from_context = self.determine_quality_answer_from_context(rag_observer)
        answer_to_question = self.determine_quality_answer_related_to_question(rag_observer)

        return AnswerQuality(answer_to_question, answer_from_context)

    def determine_quality_answer_related_to_question(self, rag_observer: RAGObserver) -> AnswerToQuestionQuality:
        chat_prompt = ChatPrompt(system_message=self.quality_of_answer_to_question_system_prompt(),
                                 user_message=self.quality_of_answer_to_question_user_prompt())

        completion = self.obtain_answer_to_question_quality(chat_prompt, rag_observer)
        quality, reason = self.split_quality_and_reason(completion)
        return AnswerToQuestionQuality(quality, reason)

    @abstractmethod
    def obtain_answer_to_question_quality(self, chat_prompt: ChatPrompt, rag_observer: RAGObserver):
        pass

    def determine_quality_answer_from_context(self, rag_observer: RAGObserver) -> AnswerFromContextQuality:
        chat_prompt = ChatPrompt(system_message=self.quality_of_answer_from_context_system_prompt(),
                                 user_message=self.quality_of_answer_from_context_user_prompt())

        completion = self.obtain_answer_from_context_quality(chat_prompt, rag_observer)
        quality, reason = self.split_quality_and_reason(completion)
        return AnswerFromContextQuality(quality, reason)

    @abstractmethod
    def obtain_answer_from_context_quality(self, chat_prompt: ChatPrompt, rag_observer: RAGObserver):
        pass

    @staticmethod
    def split_quality_and_reason(quality_and_reason: str) -> (int, str):
        split = quality_and_reason.split("-", 1)
        quality = int(split[0].strip())
        reason = split[1].strip()
        return quality, reason

    @staticmethod
    def quality_of_answer_to_question_system_prompt():
        return ("You are a quality assistant verifying retrieval augmented generation systems. Your task is to "
                "verify a generated answer against the proposed question. Give the answer a score between 1 and 5 "
                "and keep the number as an integer. 5 means the answer contains the answer to the proposed question "
                "completely. 1 means there is not match between the answer and the question at all. The question "
                "provided after 'question:'. The answer after 'answer:'. Write your answers in the format of score - "
                "reason. Keep the reason short as in maximum 2 sentences. An example: 3 - The answer is correct but "
                "some details are missing.")

    @staticmethod
    def quality_of_answer_from_context_system_prompt():
        return ("You are a quality assistant verifying retrieval augmented generation systems. Your task is to verify "
                "a generated answer against the provided context. Give the answer a score between 1 and 5 and keep "
                "the number as an integer. 5 means the answer contains only facts from the context. 1 means there is "
                "not match between the answer and the provided context at all. If the answer contains exact phrases "
                "from the context, the score should be lower as well. The answer provided after 'answer:'. The "
                "context after 'context:'. Write your answers in the format of score - reason. Keep the reason short "
                "as in maximum 2 sentences. An example: 3 - The answer is correct but some details are missing.")

    @staticmethod
    def quality_of_answer_to_question_user_prompt():
        return "Question: {question}\nAnswer: {answer}\nResult:\n"

    @staticmethod
    def quality_of_answer_from_context_user_prompt():
        return "Answer: {answer}\nContext: {context}\nResult:\n"



from rag4p.rag.generation.answer_generator import AnswerGenerator
from rag4p.rag.tracker.rag_tracker import global_data


class ObservedAnswerGenerator(AnswerGenerator):
    def __init__(self, answer_generator: AnswerGenerator):
        self.answer_generator = answer_generator

    def generate_answer(self, question: str, context: str) -> str:
        answer = self.answer_generator.generate_answer(question, context)
        global_data["observer"].question = question
        global_data["observer"].context = context
        global_data["observer"].answer = answer
        return answer



from typing import List

from rag4p.integrations.weaviate import COLLECTION_NAME
from rag4p.integrations.weaviate.access_weaviate import AccessWeaviate
from rag4p.rag.model.chunk import Chunk
from rag4p.rag.store.content_store import ContentStore
from rag4p.rag.embedding.embedder import Embedder


class WeaviateContentStore(ContentStore):

    def __init__(self, weaviate_access: AccessWeaviate, embedder: Embedder, collection_name: str = COLLECTION_NAME):
        super().__init__({
            'name': 'weaviate-content-store',
            'embedder': embedder.identifier(),
            'collection_name': collection_name,
        })
        self.embedder = embedder
        self.weaviate_access = weaviate_access
        self.collection_name = collection_name

    def store(self, chunks: List[Chunk]):
        for chunk in chunks:
            properties = {
                "documentId": chunk.document_id,
                "chunkId": chunk.chunk_id,
                "text": chunk.chunk_text,
                "totalChunks": len(chunks),
            }

            for key, value in chunk.properties.items():
                properties[key] = value

            self.weaviate_access.add_document(
                collection_name=self.collection_name,
                properties=properties,
                vector=self.embedder.embed(chunk.chunk_text)
            )



from rag4p.indexing.input_document import InputDocument
from rag4p.indexing.splitter import Splitter
from rag4p.rag.model.chunk import Chunk
from nltk.tokenize import sent_tokenize
from typing import List


class SentenceSplitter(Splitter):
    """
    Splits an InputDocument into Chunks of a single sentence.
    """

    def split(self, input_document: InputDocument) -> List[Chunk]:
        sentences = sent_tokenize(input_document.text)
        chunks = []
        for i in range(len(sentences)):
            chunk = Chunk(input_document.document_id, i, len(sentences), sentences[i], input_document.properties)
            chunks.append(chunk)
        return chunks

    @staticmethod
    def name() -> str:
        return "SentenceSplitter"



import time

from abc import ABC

from rag4p.indexing.content_reader import ContentReader
from rag4p.indexing.indexing_response import IndexingResponse
from rag4p.indexing.input_document import InputDocument
from rag4p.rag.store.content_store import ContentStore
from rag4p.indexing.splitter import Splitter


class IndexingService(ABC):
    def __init__(self, content_store: ContentStore):
        self.content_store = content_store

    def index_documents(self, content_reader: ContentReader, splitter: Splitter) -> IndexingResponse:
        start_time = time.time()
        num_documents = 0
        num_chunks = 0
        for batch in content_reader.read():
            num_documents += len(batch)
            print(f"Indexing batch of size {len(batch)}")
            for document in batch:
                doc_chunks = self.index_document(document, splitter)
                num_chunks += doc_chunks

        end_time = time.time()
        response_time = end_time - start_time

        return IndexingResponse(
            num_documents=num_documents,
            num_chunks=num_chunks,
            content_reader=content_reader.name(),
            splitter=splitter.name(),
            running_time=response_time
        )

    def index_document(self, document: InputDocument, splitter: Splitter) -> int:
        chunks = splitter.split(document)
        self.content_store.store(chunks)
        return len(chunks)


from rag4p.integrations.ollama import EMBEDDING_MODEL_NOMIC, EMBEDDING_MODEL_NOMIC_TOKENIZER, EMBEDDING_MODEL_MINILM, \
    EMBEDDING_MODEL_MINILM_TOKENIZER


def tokenizer_for_model(embedder_model_name: str):
    """
    Returns the tokenizer for the given embedder model name. First check the available models in the ollama module.
    Then look for the Hugging Face models for that specific model. Return the tokenizer to be used.
    :param embedder_model_name:
    :return:
    """
    if embedder_model_name == EMBEDDING_MODEL_NOMIC:
        return EMBEDDING_MODEL_NOMIC_TOKENIZER
    elif embedder_model_name == EMBEDDING_MODEL_MINILM:
        return EMBEDDING_MODEL_MINILM_TOKENIZER
    else:
        raise ValueError(f"Unknown embedder model name: {embedder_model_name}")


from abc import ABC, abstractmethod


class Embedder(ABC):
    """
    Abstract class for embedding text.
    """
    @abstractmethod
    def embed(self, text: str) -> [float]:
        pass

    @abstractmethod
    def identifier(self) -> str:
        pass

    @staticmethod
    def supplier() -> str:
        pass

    @abstractmethod
    def model(self) -> str:
        pass




PROVIDER = "ollama"

EMBEDDING_MODEL_NOMIC = "nomic-embed-text"
EMBEDDING_MODEL_NOMIC_TOKENIZER = "nomic-ai/nomic-embed-text-v1.5"
EMBEDDING_MODEL_MINILM = "all-minilm"
EMBEDDING_MODEL_MINILM_TOKENIZER = "sentence-transformers/all-MiniLM-L6-v2"

MODEL_PHI3 = "phi3"
MODEL_LLAMA3 = "llama3"

DEFAULT_MODEL = MODEL_PHI3
DEFAULT_EMBEDDING_MODEL = EMBEDDING_MODEL_NOMIC



from opensearchpy import OpenSearch, RequestsHttpConnection, AWSV4SignerAuth
from requests_aws4auth import AWS4Auth
import boto3


def build_local_docker_connection() -> OpenSearch:
    """
    Build a connection to a local docker container running OpenSearch.
    """
    return OpenSearch(
        hosts=[{'host': 'localhost', 'port': '9200'}],
        use_ssl=True,
        verify_certs=False,
        http_auth=('admin', 'admin'),
        ssl_assert_hostname=False,
        ssl_show_warn=False
    )


def build_aws_search_service(stack_name: str, application_prefix: str) -> OpenSearch:
    """
    Build a connection to an AWS OpenSearch service. This part expects to use the setup according to my CDK stack.
    """
    session = boto3.Session(profile_name='default')
    cfn = session.client('cloudformation')
    response = cfn.describe_stacks(
        StackName=stack_name
    )
    outputs = {
        output['ExportName']: output['OutputValue']
        for output in response['Stacks'][0]['Outputs']
        if 'ExportName' in output
    }
    host = 'https://' + outputs[f'{application_prefix}-DomainEndpoint']
    port = 443
    region = 'eu-west-1'  # e.g. us-west-1

    sts = session.client('sts')
    response = sts.assume_role(
        RoleArn=outputs[f'{application_prefix}-AdminUserRoleArn'],
        RoleSessionName="assumed-opensearch-user-admin-role",
    )
    auth = AWS4Auth(
        response['Credentials']['AccessKeyId'], response['Credentials']['SecretAccessKey'],
        region, 'es',
        session_token=response['Credentials']['SessionToken']
    )
    return OpenSearch(
        hosts=[f'{host}:{port}'],
        http_auth=auth,
        use_ssl=True,
        verify_certs=True,
        connection_class=RequestsHttpConnection
    )



from rag4p.integrations.bedrock import MODEL_TITAN_EXPRESS
from rag4p.integrations.bedrock.access_bedrock import AccessBedrock
from rag4p.rag.generation.chat.chat_prompt import ChatPrompt
from rag4p.rag.generation.quality.answer_quality_service import AnswerQualityService
from rag4p.rag.tracker.rag_observer import RAGObserver


class BedrockAnswerQualityService(AnswerQualityService):

    def __init__(self, access_bedrock: AccessBedrock):
        self.bedrock = access_bedrock

    def obtain_answer_to_question_quality(self, chat_prompt: ChatPrompt, rag_observer: RAGObserver):
        prompt_user = chat_prompt.create_user_message(params={
            "context": rag_observer.context,
            "answer": rag_observer.answer
        })
        prompt_system = chat_prompt.create_system_message(params={})
        prompt = prompt_system + prompt_user
        return self.bedrock.generate_answer(prompt=prompt, model=MODEL_TITAN_EXPRESS)

    def obtain_answer_from_context_quality(self, chat_prompt: ChatPrompt, rag_observer: RAGObserver):
        prompt_user = chat_prompt.create_user_message(params={
            "context": rag_observer.context,
            "answer": rag_observer.answer
        })
        prompt_system = chat_prompt.create_system_message(params={})
        prompt = prompt_system + prompt_user
        return self.bedrock.generate_answer(prompt=prompt, model=MODEL_TITAN_EXPRESS)



import json
from abc import abstractmethod
from typing import Iterable, List

from rag4p.indexing.content_reader import ContentReader
from rag4p.indexing.input_document import InputDocument


class JsonlContentReader(ContentReader):
    def __init__(self, file_path: str):
        self.file_path = file_path

    def read(self, batch_size: int = 10) -> Iterable[List[InputDocument]]:
        with open(self.file_path, 'r') as file:
            batch = []
            for line in file:
                data = json.loads(line)
                document = self.map_to_input_document(data)
                batch.append(document)
                if len(batch) == batch_size:
                    yield batch
                    batch = []
            if batch:
                yield batch

    @abstractmethod
    def map_to_input_document(self, data) -> InputDocument:
        pass



from abc import ABC, abstractmethod

from rag4p.rag.retrieval.retrieval_output import RetrievalOutput


class RetrievalStrategy(ABC):

    def retrieve(self, question: str) -> RetrievalOutput:
        return self.retrieve_max_results(question, 4)

    @abstractmethod
    def retrieve_max_results(self, question: str, max_results: int) -> RetrievalOutput:
        pass

    @abstractmethod
    def retrieve_max_results_observed(self, question: str, max_results: int) -> RetrievalOutput:
        pass



from abc import ABC, abstractmethod
from typing import List, Iterable

from rag4p.indexing.input_document import InputDocument


class ContentReader(ABC):
    @abstractmethod
    def read(self, batch_size: int = 10) -> Iterable[List[InputDocument]]:
        """
        Read input documents in batches.

        Parameters:
        batch_size (int): The number of documents in each chunk.

        Returns:
        Iterable[List[InputDocument]]: An iterable of lists of input documents.
        """
        pass

    def name(self) -> str:
        return self.__class__.__name__



from rag4p.rag.model.chunk import Chunk


class RelevantChunk (Chunk):
    score: float

    def __init__(self, document_id, chunk_id, total_chunks, text, properties, score):
        super().__init__(document_id, chunk_id, total_chunks, text, properties)
        self.score = score



import numpy as np
import onnxruntime as ort

from typing import List
from rag4p.rag.embedding.embedder import Embedder
from tokenizers import Tokenizer


class OnnxEmbedder(Embedder):
    """
    Embedder that uses a ONNX model to embed text. The model is a MiniLM model that has been converted to ONNX. The
    model is loaded from the file all-minilm-l6-v2-q.onnx. The tokenizer is loaded from the file tokenizer.json.
    The model runs on your local machine.
    """

    def __init__(self, max_length: int = 512):
        self.max_length = max_length
        self.tokenizer = Tokenizer.from_file("../data/tokenizer.json")
        self.tokenizer.enable_truncation(max_length=max_length)
        self.tokenizer.enable_padding(pad_to_multiple_of=1)
        self.ort_sess = ort.InferenceSession('../data/all-minilm-l6-v2-q.onnx')

    def model(self) -> str:
        return "MiniLM"

    def identifier(self) -> str:
        return "onnx-embedder-mini-lm"

    def embed(self, text: str) -> List[float]:
        tokens = self.tokenizer.encode(text, add_special_tokens=True)

        # Inference
        token_embeddings = self.ort_sess.run(None, {
            'input_ids': (np.array([tokens.ids])),
            'attention_mask': np.array([tokens.attention_mask]),
            'token_type_ids': (np.array([tokens.type_ids])),
        })
        token_embeddings = np.array(token_embeddings)

        # Mean Pooling (note: assumes that there are no padded elements)
        embeddings = np.mean(token_embeddings, axis=2)
        return embeddings[0][0]

    @staticmethod
    def supplier() -> str:
        return "ONNX"



from openai import OpenAI

from rag4p.integrations.openai import EMBEDDING_SMALL
from rag4p.rag.embedding.embedder import Embedder


class OpenAIEmbedder(Embedder):
    """
    Embedder that uses OpenAI's API to embed text. Look at the documentation for more information about the models,
    https://platform.openai.com/docs/models/embeddings. At the moment the default model is text-embedding-3-small.
    Before that text-embedding-ada-002 was used.
    """

    def __init__(self, api_key: str, embedding_model: str = EMBEDDING_SMALL):
        self.client = OpenAI(
            # This is the default and can be omitted
            api_key=api_key,
        )

        self.embedding_model = embedding_model

    def model(self) -> str:
        return self.embedding_model

    def identifier(self) -> str:
        return f"{self.supplier().lower()}-embedder-{self.embedding_model.lower()}"

    def embed(self, text: str) -> [float]:
        response = self.client.embeddings.create(input=text, model=self.embedding_model, encoding_format="float")
        embeddings = response.data

        return embeddings[0].embedding

    @staticmethod
    def supplier() -> str:
        return "OpenAI"



from rag4p.rag.model.chunk import Chunk
from rag4p.indexing.input_document import InputDocument
from rag4p.indexing.splitter import Splitter


class SingleChunkSplitter(Splitter):
    """
    Does not really split the document, it results in one chunk with the whole document.
    """

    def split(self, input_document: InputDocument) -> [str]:
        chunk = Chunk(input_document.document_id, 1, 1, input_document.text, input_document.properties)
        return [chunk]

    @staticmethod
    def name() -> str:
        return "SingleChunkSplitter"



from typing import List

from rag4p.integrations.opensearch import DEFAULT_INDEX
from rag4p.integrations.opensearch.opensearch_client import OpenSearchClient
from rag4p.rag.embedding.embedder import Embedder
from rag4p.rag.model.chunk import Chunk
from rag4p.rag.store.content_store import ContentStore


class OpenSearchContentStore(ContentStore):

    def __init__(self, opensearch_client: OpenSearchClient, embedder: Embedder, index_name: str = DEFAULT_INDEX):
        super().__init__({
            'name': 'opensearch-content-store',
            'embedder': embedder.identifier(),
            'index_name': index_name,
        })

        self.embedder = embedder
        self.index_name = index_name
        self.client = opensearch_client
        self.client.ping()

    def store(self, chunks: List[Chunk]):
        for chunk in chunks:
            properties = {
                "document_id": chunk.document_id,
                "chunk_id": chunk.chunk_id,
                "chunk_text": chunk.chunk_text,
                "total_chunks": len(chunks),
                "chunk_vector": self.embedder.embed(chunk.chunk_text),
            }

            for key, value in chunk.properties.items():
                properties[key] = value

            self.client.index_document(
                id=chunk.get_id(),
                document=properties,
                index_name=self.index_name,
            )



from datetime import datetime
import json
import pickle
from typing import List

import pandas as pd

from rag4p.rag.model.chunk import Chunk
from rag4p.rag.model.relevant_chunk import RelevantChunk
from rag4p.rag.store.content_store import ContentStore
from rag4p.rag.embedding.embedder import Embedder
from rag4p.rag.retrieval.retriever import Retriever

from scipy.spatial import distance


class InternalContentStore(ContentStore, Retriever):
    """
    The internal content stores stores the chunks in memory, it acts as a normal content store, but it als contains
    all the methods from a retriever, so it can be used as a retriever as well. This is useful for testing purposes.
    """

    def __init__(self, embedder: Embedder, metadata=None):
        _metadata = {
                'name': 'internal-content-store',
                'create_date': datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                'embedder': embedder.identifier(),
                'supplier': embedder.supplier(),
                'model': embedder.model()
            }

        if metadata is not None:
            _metadata = {**_metadata, **metadata}

        super().__init__(_metadata)
        self.embedder = embedder
        self.vector_store = pd.DataFrame(columns=['chunk_id', 'chunk', 'embedding'])

    def store(self, chunks: List[Chunk]):
        for chunk in chunks:
            self.__store_chunk(chunk)

    def __store_chunk(self, chunk: Chunk):
        chunk_id = chunk.document_id + "_" + str(chunk.chunk_id)
        print(f"Storing chunk {chunk_id}: {chunk.chunk_text}")
        embedding = self.embedder.embed(chunk.chunk_text)
        self.vector_store.loc[len(self.vector_store)] = {'chunk_id': chunk_id, 'chunk': chunk, 'embedding': embedding}

    def find_relevant_chunks(self, query: str, max_results: int = 4) -> List[RelevantChunk]:
        print(f"Finding relevant chunks for query: {query}")
        embedding = self.embedder.embed(query)
        self.vector_store['distance'] = self.vector_store['embedding'].apply(lambda x: distance.euclidean(x, embedding))
        relevant_chunks_df = self.vector_store.nsmallest(max_results, 'distance')

        relevant_chunks = []
        for index, row in relevant_chunks_df.iterrows():
            chunk = row['chunk']
            score = row['distance']
            relevant_chunk = RelevantChunk(
                document_id=chunk.document_id,
                chunk_id=chunk.chunk_id,
                total_chunks=chunk.total_chunks,
                text=chunk.chunk_text,
                properties=chunk.properties,
                score=score
            )
            relevant_chunks.append(relevant_chunk)
        return relevant_chunks

    def get_chunk_by_id(self, chunk_id: str) -> Chunk:
        """
        Obtains a chunk using its complete id (document_id + "_" + chunk_id)
        :param chunk_id: complete id of the chunk document_id + "_" + chunk_id
        :return:
        """
        found_chunk = self.vector_store[self.vector_store['chunk_id'] == chunk_id]

        if len(found_chunk) == 0:
            raise Exception(f"Chunk with id {chunk_id} not found.")

        return found_chunk.iloc[0]['chunk']

    def loop_over_chunks(self):
        for index, row in self.vector_store.iterrows():
            yield row['chunk']

    def backup(self, path: str):
        # Save the DataFrame to a pickle file
        with open(f'{path}.pickle', 'wb') as f:
            pickle.dump(self.vector_store, f)

        # Save the metadata to a JSON file
        with open(f'{path}_metadata.json', 'w') as f:
            json.dump(self._metadata, f)

    @classmethod
    def load_from_backup(cls, embedder: Embedder, path: str):
        # Create an instance of the class
        instance = cls(embedder)

        # Load the DataFrame from the pickle file
        with open(f'{path}.pickle', 'rb') as f:
            instance.vector_store = pickle.load(f)

        # Load the metadata from the JSON file
        with open(f'{path}_metadata.json', 'r') as f:
            instance._metadata = json.load(f)

        if 'embedder' in instance._metadata:
            if instance._metadata['embedder'] != embedder.identifier():
                raise Exception(f"Embedder {embedder.identifier()} does not match the one in the backup: "
                                f"{instance._metadata['embedder']}")

        return instance



class RAGObserver:
    question: str
    answer: str
    context: str

    def __init__(self, question: str = "", answer: str = "", context: str = ""):
        self.question = question
        self.answer = answer
        self.context = context
        self.relevant_chunks = {}
        self.relevant_chunks_with_windows = {}
        self.relevant_chunks_windows_text = {}

    def add_relevant_chunk(self, chunk_id: str, text: str):
        """Add a relevant chunk to the observer. Chunk id is the document id + "_" + chunk id."""
        self.relevant_chunks[chunk_id] = text

    def add_window_to_relevant_chunk(self, chunk_id: str, window: [int]):
        """Add a window to a relevant chunk. Chunk id is the document id + "_" + chunk id."""
        self.relevant_chunks[chunk_id] = window

    def add_window_text_to_relevant_chunk(self, chunk_id: str, window_text: str):
        """Add a window text to a relevant chunk. Chunk id is the document id + "_" + chunk id."""
        self.relevant_chunks_windows_text[chunk_id] = window_text

    def reset(self):
        """Reset the observer."""
        self.question = ""
        self.answer = ""
        self.context = ""
        self.relevant_chunks = {}
        self.relevant_chunks_with_windows = {}
        self.relevant_chunks_windows_text = {}


from dotenv import load_dotenv

from rag4p.indexing.indexing_service import IndexingService
from rag4p.indexing.splitters.sentence_splitter import SentenceSplitter
from rag4p.integrations.openai.openai_embedder import OpenAIEmbedder
from rag4p.integrations.opensearch.connection_builder import build_aws_search_service
from rag4p.integrations.opensearch.index_components import ComponentTemplate, ComponentSettings, \
    ComponentDynamicMappings, ComponentMappings
from rag4p.integrations.opensearch.opensearch_client import OpenSearchClient
from rag4p.integrations.opensearch.opensearch_content_store import OpenSearchContentStore
from rag4p.integrations.opensearch.opensearch_template import OpenSearchTemplate
from rag4p.util.key_loader import KeyLoader
from rag4p.vasa_content_reader import VasaContentReader

if __name__ == '__main__':
    load_dotenv()
    key_loader = KeyLoader()

    opensearch_conn = build_aws_search_service(stack_name=key_loader.get_property("OPENSEARCH_STACK_NAME"),
                                                 application_prefix=key_loader.get_property("OPENSEARCH_APP_PREFIX"))

    opensearch_client = OpenSearchClient(opensearch_conn)

    index_template = ComponentTemplate(name="vasa_index_template",
                                       version=2,
                                       index_name="rag4p-vasa",
                                       component_names=["vasa_settings", "vasa_mappings", "vasa_dynamic_mappings"],
                                       embedding_dimension=1536)
    component_settings = ComponentSettings(name="vasa_settings", version=1, settings={})
    component_dyn_mappings = ComponentDynamicMappings(name="vasa_dynamic_mappings", version=1, dynamic_mappings=[])
    component_mappings = ComponentMappings(name="vasa_mappings", version=1, mappings={
        "title": {
            "type": "text"
        },
        "timerange": {
            "type": "text"
        },
    })
    opensearch_template = OpenSearchTemplate(client=opensearch_client,
                                             index_template=index_template,
                                             component_settings=component_settings,
                                             component_dyn_mappings=component_dyn_mappings,
                                             component_mappings=component_mappings)

    opensearch_template.create_update_template()
    index_name = opensearch_client.create_index(provided_alias_name="rag4p-vasa")

    embedder = OpenAIEmbedder(api_key=key_loader.get_openai_api_key())
    opensearch_content_store = OpenSearchContentStore(opensearch_client=opensearch_client,
                                                      embedder=embedder,
                                                      index_name=index_name)

    splitter = SentenceSplitter()
    indexing_service = IndexingService(content_store=opensearch_content_store)

    content_reader = VasaContentReader()
    indexing_service.index_documents(content_reader=content_reader, splitter=splitter)

    opensearch_client.switch_alias_to(index_name=index_name, provided_alias_name="rag4p-vasa")



from rag4p.indexing.input_document import InputDocument
from rag4p.indexing.jsonl_content_reader import JsonlContentReader


class VasaContentReader(JsonlContentReader):
    def __init__(self):
        super().__init__(file_path="../data/vasa-timeline.jsonl")

    def map_to_input_document(self, data) -> InputDocument:
        properties = {
            "title": data["title"],
            "timerange": data["timerange"]
        }
        document_id = data["title"].lower().replace(" ", "-")
        document = InputDocument(
            document_id=document_id,
            text=data["body"],
            properties=properties
        )
        return document



from rag4p.integrations.ollama import DEFAULT_MODEL
from rag4p.integrations.ollama.access_ollama import AccessOllama
from rag4p.rag.generation.answer_generator import AnswerGenerator


class OllamaAnswerGenerator(AnswerGenerator):

    def __init__(self, access_ollama: AccessOllama, model: str = DEFAULT_MODEL):
        self.ollama = access_ollama
        self.model = model

    def generate_answer(self, question: str, context: str) -> str:
        prompt = f"""
        You are an assistant answering questions using the context provided. If the context does not contain the
        answer, you should tell you cannot answer using the context.
        question: {question}
        context: {context}
        answer:
        """
        return self.ollama.generate_answer(prompt=prompt, model=self.model)



"""
This model provides access to the OpenAI API. It is used to embed text and to generate answers to questions. We also
provide a default model for both embedding and answering.
"""

PROVIDER = "openai"

MODEL_GPT4O = "gpt-4o"
MODEL_GPT4 = "gpt-4"
MODEL_GPT4_TURBO = "gpt-4-turbo"
MODEL_GPT35_TURBO = "gpt-3.5-turbo"

EMBEDDING_ADA = "text-embedding-ada-002"
EMBEDDING_SMALL = "text-embedding-3-small"

DEFAULT_MODEL = MODEL_GPT4O
DEFAULT_EMBEDDING_MODEL = EMBEDDING_SMALL



import csv
import os

from rag4p.rag.generation.question_generator import QuestionGenerator
from rag4p.rag.retrieval.retriever import Retriever


class QuestionGeneratorService:

    def __init__(self, retriever: Retriever, question_generator: QuestionGenerator):
        self.retriever = retriever
        self.question_generator = question_generator

    def generate_question_answer_pairs(self, file_name: str):
        directory = os.getcwd()  # get current working directory
        file_path = os.path.join(directory, "../data", file_name)
        print(file_path)
        try:
            with open(file_path, 'w', newline='', encoding='utf-8') as file:
                writer = csv.writer(file)
                writer.writerow(["document", "chunk", "text", "question"])
                for chunk in self.retriever.loop_over_chunks():
                    # Use LLM to generate a question for this chunk
                    question = self.generate_question(chunk.chunk_text)

                    # Write the question and answer to a file
                    document_id = chunk.document_id
                    chunk_id = str(chunk.chunk_id)
                    writer.writerow([document_id, chunk_id, chunk.chunk_text, question])
                    print(f"Generated question: {question}")
        except IOError as e:
            print("An error occurred while writing to the file.", e)
            raise

    def generate_question(self, context: str) -> str:
        """
        Generates a question from a context.

        :param context: The context to generate a question from.
        :return: The generated question.
        """
        return self.question_generator.generate_question(context)


from abc import abstractmethod, ABC
from typing import List

from rag4p.rag.model.chunk import Chunk
from rag4p.indexing.input_document import InputDocument


class Splitter(ABC):

    @abstractmethod
    def split(self, input_document: InputDocument) -> List[Chunk]:
        pass

    @staticmethod
    @abstractmethod
    def name() -> str:
        pass
    


from abc import ABC, abstractmethod

from rag4p.rag.model.chunk import Chunk
from rag4p.rag.model.relevant_chunk import RelevantChunk


class Retriever(ABC):
    """
    This interface is used to retrieve relevant chunks for a given question. Next to the functions to retrieve the
    answer using the question or the vector representation of the question, it also provides a function to loop over
    all chunks. Finally, it contains a method to get a specific chunk.
    """

    @abstractmethod
    def find_relevant_chunks(self, question: str, max_results: int = 4) -> [RelevantChunk]:
        pass

    def get_chunk(self, document_id: str, chunk_id: int) -> Chunk:
        return self.get_chunk_by_id(document_id + "_" + str(chunk_id))

    @abstractmethod
    def get_chunk_by_id(self, chunk_id: str) -> Chunk:
        pass

    @abstractmethod
    def loop_over_chunks(self):
        pass



from rag4p.indexing.indexing_service import IndexingService
from rag4p.indexing.splitters.single_chunk_splitter import SingleChunkSplitter
from rag4p.rag.embedding.local.onnx_embedder import OnnxEmbedder
from rag4p.rag.store.local.internal_content_store import InternalContentStore
from rag4p.vasa_content_reader import VasaContentReader


if __name__ == '__main__':
    from dotenv import load_dotenv
    load_dotenv()

    example_sentences = [
        "How many bolts were replaced?",
        "Since When could people visit the Vasa?",
        "Since when was the Vasa available for the public to visit?",
        "Who was responsible for building the Vasa ship?",
        "Where did the person responsible for building the Vasa ship come from?"
    ]

    splitter = SingleChunkSplitter()
    embedder = OnnxEmbedder()
    content_store = InternalContentStore(embedder=embedder)

    indexing_service = IndexingService(content_store=content_store)
    indexing_service.index_documents(content_reader=VasaContentReader(), splitter=splitter)

    for example_sentence in example_sentences:
        relevant_chunks = content_store.find_relevant_chunks(example_sentence, max_results=2)
        print(f"Found {len(relevant_chunks)} relevant chunks for query: {example_sentence}")
        for chunk in relevant_chunks:
            print(f"Document: {chunk.document_id}")
            print(f"Chunk id: {chunk.chunk_id}")
            print(f"Text: {chunk.chunk_text}")
            print(f"Distance: {chunk.score:.3f}")


from abc import ABC, abstractmethod
from types import MappingProxyType
from typing import List

from rag4p.rag.model.chunk import Chunk


class ContentStore(ABC):

    def __init__(self, metadata=None):
        if metadata is None:
            metadata = {}
        self._metadata = metadata

    def get_metadata(self):
        return MappingProxyType(self._metadata)

    def add_metadata(self, key: str, value):
        self._metadata[key] = value

    @abstractmethod
    def store(self, chunks: List[Chunk]):
        pass



from rag4p.indexing.splitters.sentence_splitter import SentenceSplitter
from rag4p.indexing.indexing_service import IndexingService
from rag4p.rag.embedding.local.onnx_embedder import OnnxEmbedder
from rag4p.rag.retrieval.quality.retrieval_quality_service import read_question_answers_from_file, \
    obtain_retrieval_quality
from rag4p.rag.store.local.internal_content_store import InternalContentStore
from rag4p.util.key_loader import KeyLoader
from rag4p.vasa_content_reader import VasaContentReader

if __name__ == '__main__':
    from dotenv import load_dotenv
    load_dotenv()

    key_loader = KeyLoader()
    embedder = OnnxEmbedder()
    content_store = InternalContentStore(embedder=embedder)
    indexing_service = IndexingService(content_store=content_store)
    indexing_service.index_documents(content_reader=VasaContentReader(), splitter=SentenceSplitter())

    question_answer_records = read_question_answers_from_file()
    retriever_quality = obtain_retrieval_quality(question_answer_records=question_answer_records,
                                                 retriever=content_store)

    print(f"Quality using precision: {retriever_quality.precision()}")
    print(f"Total questions: {retriever_quality.total_questions()}")



import weaviate.classes.config as wvc


def weaviate_properties(additional_properties=None):

    if additional_properties is None:
        additional_properties = []

    common_properties = [
        wvc.Property(name="documentId",
                     data_type=wvc.DataType.TEXT,
                     vectorize_property_name=False,
                     skip_vectorization=True),
        wvc.Property(name="chunkId",
                     data_type=wvc.DataType.INT,
                     vectorize_property_name=False,
                     skip_vectorization=True),
        wvc.Property(name="totalChunks",
                     data_type=wvc.DataType.INT,
                     vectorize_property_name=False,
                     skip_vectorization=True),
        wvc.Property(name="text",
                     data_type=wvc.DataType.TEXT,
                     vectorize_property_name=False,
                     skip_vectorization=False),
    ]

    return common_properties + additional_properties



from abc import ABC, abstractmethod
from typing import List


class IndexComponent(ABC):
    def __init__(self, name: str, version: int):
        self.name = name
        self.version = version

    @abstractmethod
    def get_body(self):
        pass


class ComponentSettings(IndexComponent):
    def __init__(self, name: str, version: int, settings: dict):
        super().__init__(name, version)
        self.settings = {
            "template": {
                "settings": settings
            },
            "version": version,
        }

    def get_body(self):
        return self.settings


class ComponentMappings(IndexComponent):
    def __init__(self, name: str, version: int, mappings: dict):
        super().__init__(name, version)
        self.mappings = {
            "template": {
                "mappings": {
                    "properties": mappings
                }
            },
            "version": version,
        }

    def get_body(self):
        return self.mappings


class ComponentDynamicMappings(IndexComponent):
    def __init__(self, name: str, version: int, dynamic_mappings: List[dict]):
        super().__init__(name, version)
        if not dynamic_mappings or len(dynamic_mappings) == 0:
            dynamic_mappings = [{
                "strings": {
                    "match_mapping_type": "string",
                    "mapping": {
                        "type": "text",
                        "fields": {
                            "keyword": {
                                "type": "keyword",
                                "ignore_above": 256
                            }
                        }
                    }
                }
            }]

        self.dynamic_mappings = {
            "template": {
                "mappings": {
                    "dynamic_templates": dynamic_mappings
                }
            },
            "version": version,
        }

    def get_body(self):
        return self.dynamic_mappings


class ComponentTemplate(IndexComponent):
    def __init__(self, name: str, version: int, index_name: str, component_names: list, embedding_dimension: int = 1536):
        super().__init__(name, version)
        self.template = {
            "index_patterns": [f"{index_name}-*"],
            "priority": 10,
            "template": {
                "mappings": {
                    "properties": {
                        "document_id": {
                            "type": "keyword"
                        },
                        "chunk_id": {
                            "type": "integer"
                        },
                        "total_chunks": {
                            "type": "integer"
                        },
                        "chunk_text": {
                            "type": "text"
                        },
                        "chunk_vector": {
                            "type": "knn_vector",
                            "dimension": embedding_dimension
                        },
                    }
                },
                "settings": {
                    "number_of_shards": 1,
                    "number_of_replicas": 0,
                    "index": {
                        "knn": "true",
                        "knn.algo_param.ef_search": 8,
                    }
                }
            },
            "version": version,
            "composed_of": component_names
        }

    def get_body(self):
        return self.template



from rag4p.rag.model.chunk import Chunk
from rag4p.rag.model.relevant_chunk import RelevantChunk
from rag4p.rag.retrieval.retriever import Retriever
from rag4p.rag.tracker.rag_tracker import global_data


class ObservedRetriever(Retriever):
    def __init__(self, retriever: Retriever):
        self.retriever = retriever

    def find_relevant_chunks(self, question: str, max_results: int = 4) -> [RelevantChunk]:
        relevant_chunks = self.retriever.find_relevant_chunks(question, max_results)
        for relevant_chunk in relevant_chunks:
            global_data["observer"].add_relevant_chunk(relevant_chunk.get_id(), relevant_chunk.text)
        return relevant_chunks

    def get_chunk(self, document_id: str, chunk_id: int) -> Chunk:
        chunk = self.retriever.get_chunk(document_id, chunk_id)
        return chunk

    def loop_over_chunks(self):
        yield from self.retriever.loop_over_chunks()


import numpy as np

from rag4p.integrations.openai.openai_answer_generator import OpenaiAnswerGenerator
from rag4p.integrations.openai.openai_embedder import OpenAIEmbedder
from rag4p.integrations.weaviate.access_weaviate import AccessWeaviate
from rag4p.integrations.weaviate.weaviate_retriever import WeaviateRetriever
from rag4p.rag.generation.observed_answer_generator import ObservedAnswerGenerator
from rag4p.rag.generation.quality.answer_quality_service import AnswerQualityService
from rag4p.rag.retrieval.strategies.window_retrieval_strategy import WindowRetrievalStrategy
from rag4p.rag.tracker.rag_tracker import global_data
from rag4p.util.key_loader import KeyLoader

if __name__ == '__main__':
    from dotenv import load_dotenv
    load_dotenv()
    key_loader = KeyLoader()

    embedder = OpenAIEmbedder(api_key=key_loader.get_openai_api_key())
    weaviate_client = AccessWeaviate(url=key_loader.get_weaviate_url(), access_key=key_loader.get_weaviate_api_key())
    retriever = WeaviateRetriever(weaviate_access=weaviate_client, embedder=embedder, hybrid=False)

    openai_answer_generator = OpenaiAnswerGenerator(openai_api_key=key_loader.get_openai_api_key())
    answer_generator = ObservedAnswerGenerator(answer_generator=openai_answer_generator)

    example_sentences = [
        "How many bolts were replaced?",
        "Since When could people visit the Vasa?",
        "Since when was the Vasa available for the public to visit?",
        "Who was responsible for building the Vasa ship?",
        "Where did the person responsible for building the Vasa ship come from?"
    ]

    strategy = WindowRetrievalStrategy(retriever=retriever, window_size=1)
    answer_quality_service = AnswerQualityService(openai_api_key=key_loader.get_openai_api_key())

    answer_question_quality = []
    answer_context_quality = []
    for example_sentence in example_sentences:
        print("\n----------------------------------")
        retrieval_output = strategy.retrieve_max_results(question=example_sentence, max_results=1)

        # Generate answer using the obtained context
        answer = answer_generator.generate_answer(question=example_sentence,
                                                  context=retrieval_output.construct_context())

        print(f"Question: {example_sentence}")
        print(f"Answer: {answer}")
        print(f"Context: {retrieval_output.construct_context()}")

        rag_observer = global_data["observer"]

        quality = answer_quality_service.determine_quality_answer_related_to_question(rag_observer=rag_observer)
        answer_question_quality.append(quality.quality)
        print(f"Answer to Question Quality: {quality.quality}, Reason: {quality.reason}")

        quality = answer_quality_service.determine_quality_answer_from_context(rag_observer=rag_observer)
        answer_context_quality.append(quality.quality)
        print(f"Answer from Context Quality: {quality.quality}, Reason: {quality.reason}")

        rag_observer.reset()

    print("\n----------------------------------")
    print(f"Question Quality: {np.mean(answer_question_quality)}")
    print(f"Context Quality: {np.mean(answer_context_quality)}")

    weaviate_client.close()



from rag4p.rag.model.relevant_chunk import RelevantChunk
from rag4p.rag.retrieval.retrieval_output import RetrievalOutput, RetrievalOutputItem
from rag4p.rag.retrieval.retrieval_strategy import RetrievalStrategy
from rag4p.rag.retrieval.retriever import Retriever
from rag4p.rag.tracker.rag_tracker import global_data


class WindowRetrievalStrategy(RetrievalStrategy):

    def __init__(self, retriever: Retriever, window_size: int = 1):
        self.retriever = retriever
        self.window_size = window_size

    def retrieve_max_results(self, question, max_results) -> RetrievalOutput:
        relevant_chunks = self.retriever.find_relevant_chunks(question, max_results)

        return self.__extract_window_for_relevant_chunks(relevant_chunks)

    def retrieve_max_results_observed(self, question, max_results) -> RetrievalOutput:
        relevant_chunks = self.retriever.find_relevant_chunks(question, max_results)

        return self.__extract_window_for_relevant_chunks(relevant_chunks, observe=True)

    def __extract_window_for_relevant_chunks(self, relevant_chunks: [RelevantChunk], observe: bool = False) \
            -> RetrievalOutput:
        retrieval_output_items = []
        for relevant_chunk in relevant_chunks:
            chunk_ids = self.__chunk_ids_for_window(relevant_chunk.chunk_id,
                                                    self.window_size,
                                                    relevant_chunk.total_chunks)
            overall_text = ""
            for chunk_id in chunk_ids:
                chunk = self.retriever.get_chunk(relevant_chunk.document_id, chunk_id)
                overall_text += chunk.chunk_text + " "

            if observe:
                global_data["observer"].add_relevant_chunk(relevant_chunk.get_id(), overall_text)

            relevant_item = RetrievalOutputItem(document_id=relevant_chunk.document_id,
                                                chunk_id=relevant_chunk.chunk_id,
                                                text=overall_text)
            retrieval_output_items.append(relevant_item)
        return RetrievalOutput(retrieval_output_items)

    @staticmethod
    def __chunk_ids_for_window(chunk_id: int, window_size: int, number_of_chunks: int) -> [int]:
        # Calculate the start and end of the window
        start = max(0, chunk_id - window_size)
        end = min(number_of_chunks - 1, chunk_id + window_size)

        # Generate the list of chunk_ids
        chunk_ids = [start + i for i in range(end - start + 1)]

        return chunk_ids



from rag4p.rag.generation.quality.answer_from_context_quality import AnswerFromContextQuality
from rag4p.rag.generation.quality.answer_to_question_quality import AnswerToQuestionQuality


class AnswerQuality:
    answer_to_question: AnswerToQuestionQuality
    answer_from_context: AnswerFromContextQuality

    def __init__(self, answer_to_question: AnswerToQuestionQuality, answer_from_context: AnswerFromContextQuality):
        self.answer_to_question = answer_to_question
        self.answer_from_context = answer_from_context
